@@ -1,35 +1,37 @@
-""" Multicast DNS Service Discovery for Python, v0.14-wmcbrine
-    Copyright 2003 Paul Scott-Murphy, 2014 William McBrine
+"""Multicast DNS Service Discovery for Python, v0.14-wmcbrine
+Copyright 2003 Paul Scott-Murphy, 2014 William McBrine
 
-    This module provides a framework for the use of DNS Service Discovery
-    using IP multicast.
+This module provides a framework for the use of DNS Service Discovery
+using IP multicast.
 
-    This library is free software; you can redistribute it and/or
-    modify it under the terms of the GNU Lesser General Public
-    License as published by the Free Software Foundation; either
-    version 2.1 of the License, or (at your option) any later version.
+This library is free software; you can redistribute it and/or
+modify it under the terms of the GNU Lesser General Public
+License as published by the Free Software Foundation; either
+version 2.1 of the License, or (at your option) any later version.
 
-    This library is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
-    Lesser General Public License for more details.
+This library is distributed in the hope that it will be useful,
+but WITHOUT ANY WARRANTY; without even the implied warranty of
+MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
+Lesser General Public License for more details.
 
-    You should have received a copy of the GNU Lesser General Public
-    License along with this library; if not, write to the Free Software
-    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
-    USA
+You should have received a copy of the GNU Lesser General Public
+License along with this library; if not, write to the Free Software
+Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
+USA
 """
 
 import enum
 import logging
-from typing import Dict, List, Optional, Sequence, Tuple, Union
+from struct import Struct
+from typing import TYPE_CHECKING, Dict, List, Optional, Sequence, Tuple, Union
 
-from .._cache import DNSCache
 from .._dns import DNSPointer, DNSQuestion, DNSRecord
 from .._exceptions import NamePartTooLongException
 from .._logger import log
 from ..const import (
     _CLASS_UNIQUE,
+    _DNS_HOST_TTL,
+    _DNS_OTHER_TTL,
     _DNS_PACKET_HEADER_LEN,
     _FLAGS_QR_MASK,
     _FLAGS_QR_QUERY,
@@ -40,31 +42,55 @@ from ..const import (
 )
 from .incoming import DNSIncoming
 
+str_ = str
+float_ = float
+int_ = int
+bytes_ = bytes
+DNSQuestion_ = DNSQuestion
+DNSRecord_ = DNSRecord
+
+
+PACK_BYTE = Struct(">B").pack
+PACK_SHORT = Struct(">H").pack
+PACK_LONG = Struct(">L").pack
+
+SHORT_CACHE_MAX = 128
+
+BYTE_TABLE = tuple(PACK_BYTE(i) for i in range(256))
+SHORT_LOOKUP = tuple(PACK_SHORT(i) for i in range(SHORT_CACHE_MAX))
+LONG_LOOKUP = {i: PACK_LONG(i) for i in (_DNS_OTHER_TTL, _DNS_HOST_TTL, 0)}
+
 
 class State(enum.Enum):
     init = 0
     finished = 1
 
 
-class DNSOutgoing:
+STATE_INIT = State.init.value
+STATE_FINISHED = State.finished.value
 
+LOGGING_IS_ENABLED_FOR = log.isEnabledFor
+LOGGING_DEBUG = logging.DEBUG
+
+
+class DNSOutgoing:
     """Object representation of an outgoing packet"""
 
     __slots__ = (
-        'flags',
-        'finished',
-        'id',
-        'multicast',
-        'packets_data',
-        'names',
-        'data',
-        'size',
-        'allow_long',
-        'state',
-        'questions',
-        'answers',
-        'authorities',
-        'additionals',
+        "additionals",
+        "allow_long",
+        "answers",
+        "authorities",
+        "data",
+        "finished",
+        "flags",
+        "id",
+        "multicast",
+        "names",
+        "packets_data",
+        "questions",
+        "size",
+        "state",
     )
 
     def __init__(self, flags: int, multicast: bool = True, id_: int = 0) -> None:
@@ -80,7 +106,7 @@ class DNSOutgoing:
         self.size: int = _DNS_PACKET_HEADER_LEN
         self.allow_long: bool = True
 
-        self.state = State.init
+        self.state = STATE_INIT
 
         self.questions: List[DNSQuestion] = []
         self.answers: List[Tuple[DNSRecord, float]] = []
@@ -102,15 +128,17 @@ class DNSOutgoing:
         self.allow_long = True
 
     def __repr__(self) -> str:
-        return '<DNSOutgoing:{%s}>' % ', '.join(
-            [
-                'multicast=%s' % self.multicast,
-                'flags=%s' % self.flags,
-                'questions=%s' % self.questions,
-                'answers=%s' % self.answers,
-                'authorities=%s' % self.authorities,
-                'additionals=%s' % self.additionals,
-            ]
+        return "<DNSOutgoing:{}>".format(
+            ", ".join(
+                [
+                    f"multicast={self.multicast}",
+                    f"flags={self.flags}",
+                    f"questions={self.questions}",
+                    f"answers={self.answers}",
+                    f"authorities={self.authorities}",
+                    f"additionals={self.additionals}",
+                ]
+            )
         )
 
     def add_question(self, record: DNSQuestion) -> None:
@@ -120,11 +148,12 @@ class DNSOutgoing:
     def add_answer(self, inp: DNSIncoming, record: DNSRecord) -> None:
         """Adds an answer"""
         if not record.suppressed_by(inp):
-            self.add_answer_at_time(record, 0)
+            self.add_answer_at_time(record, 0.0)
 
-    def add_answer_at_time(self, record: Optional[DNSRecord], now: Union[float, int]) -> None:
+    def add_answer_at_time(self, record: Optional[DNSRecord], now: float_) -> None:
         """Adds an answer if it does not expire by a certain time"""
-        if record is not None and (now == 0 or not record.is_expired(now)):
+        now_double = now
+        if record is not None and (now_double == 0 or not record.is_expired(now_double)):
             self.answers.append((record, now))
 
     def add_authorative_answer(self, record: DNSPointer) -> None:
@@ -169,61 +198,48 @@ class DNSOutgoing:
         """
         self.additionals.append(record)
 
-    def add_question_or_one_cache(
-        self, cache: DNSCache, now: float, name: str, type_: int, class_: int
-    ) -> None:
-        """Add a question if it is not already cached."""
-        cached_entry = cache.get_by_details(name, type_, class_)
-        if not cached_entry:
-            self.add_question(DNSQuestion(name, type_, class_))
-        else:
-            self.add_answer_at_time(cached_entry, now)
-
-    def add_question_or_all_cache(
-        self, cache: DNSCache, now: float, name: str, type_: int, class_: int
-    ) -> None:
-        """Add a question if it is not already cached.
-        This is currently only used for IPv6 addresses.
-        """
-        cached_entries = cache.get_all_by_details(name, type_, class_)
-        if not cached_entries:
-            self.add_question(DNSQuestion(name, type_, class_))
-            return
-        for cached_entry in cached_entries:
-            self.add_answer_at_time(cached_entry, now)
-
-    def _write_byte(self, value: int) -> None:
+    def _write_byte(self, value: int_) -> None:
         """Writes a single byte to the packet"""
-        self.data.append(value.to_bytes(1, 'big'))
+        self.data.append(BYTE_TABLE[value])
         self.size += 1
 
-    def _insert_short_at_start(self, value: int) -> None:
+    def _get_short(self, value: int_) -> bytes:
+        """Convert an unsigned short to 2 bytes."""
+        return SHORT_LOOKUP[value] if value < SHORT_CACHE_MAX else PACK_SHORT(value)
+
+    def _insert_short_at_start(self, value: int_) -> None:
         """Inserts an unsigned short at the start of the packet"""
-        self.data.insert(0, value.to_bytes(2, 'big'))
+        self.data.insert(0, self._get_short(value))
 
-    def _replace_short(self, index: int, value: int) -> None:
+    def _replace_short(self, index: int_, value: int_) -> None:
         """Replaces an unsigned short in a certain position in the packet"""
-        self.data[index] = value.to_bytes(2, 'big')
+        self.data[index] = self._get_short(value)
 
-    def write_short(self, value: int) -> None:
+    def write_short(self, value: int_) -> None:
         """Writes an unsigned short to the packet"""
-        self.data.append(value.to_bytes(2, 'big'))
+        self.data.append(self._get_short(value))
         self.size += 2
 
     def _write_int(self, value: Union[float, int]) -> None:
         """Writes an unsigned integer to the packet"""
-        self.data.append(int(value).to_bytes(4, 'big'))
+        value_as_int = int(value)
+        long_bytes = LONG_LOOKUP.get(value_as_int)
+        if long_bytes is not None:
+            self.data.append(long_bytes)
+        else:
+            self.data.append(PACK_LONG(value_as_int))
         self.size += 4
 
-    def write_string(self, value: bytes) -> None:
+    def write_string(self, value: bytes_) -> None:
         """Writes a string to the packet"""
-        assert isinstance(value, bytes)
+        if TYPE_CHECKING:
+            assert isinstance(value, bytes)
         self.data.append(value)
         self.size += len(value)
 
-    def _write_utf(self, s: str) -> None:
+    def _write_utf(self, s: str_) -> None:
         """Writes a UTF-8 string of a given length to the packet"""
-        utfstr = s.encode('utf-8')
+        utfstr = s.encode("utf-8")
         length = len(utfstr)
         if length > 64:
             raise NamePartTooLongException
@@ -231,14 +247,15 @@ class DNSOutgoing:
         self.write_string(utfstr)
 
     def write_character_string(self, value: bytes) -> None:
-        assert isinstance(value, bytes)
+        if TYPE_CHECKING:
+            assert isinstance(value, bytes)
         length = len(value)
         if length > 256:
             raise NamePartTooLongException
         self._write_byte(length)
         self.write_string(value)
 
-    def write_name(self, name: str) -> None:
+    def write_name(self, name: str_) -> None:
         """
         Write names to packet
 
@@ -252,54 +269,68 @@ class DNSOutgoing:
         """
 
         # split name into each label
-        name_length = None
-        if name.endswith('.'):
-            name = name[: len(name) - 1]
-        labels = name.split('.')
-        # Write each new label or a pointer to the existing
-        # on in the packet
+        if name.endswith("."):
+            name = name[:-1]
+
+        index = self.names.get(name, 0)
+        if index:
+            self._write_link_to_name(index)
+            return
+
         start_size = self.size
-        for count in range(len(labels)):
-            label = name if count == 0 else '.'.join(labels[count:])
-            index = self.names.get(label)
+        labels = name.split(".")
+        # Write each new label or a pointer to the existing one in the packet
+        self.names[name] = start_size
+        self._write_utf(labels[0])
+
+        name_length = 0
+        for count in range(1, len(labels)):
+            partial_name = ".".join(labels[count:])
+            index = self.names.get(partial_name, 0)
             if index:
-                # If part of the name already exists in the packet,
-                # create a pointer to it
-                self._write_byte((index >> 8) | 0xC0)
-                self._write_byte(index & 0xFF)
+                self._write_link_to_name(index)
                 return
-            if name_length is None:
-                name_length = len(name.encode('utf-8'))
-            self.names[label] = start_size + name_length - len(label.encode('utf-8'))
+            if name_length == 0:
+                name_length = len(name.encode("utf-8"))
+            self.names[partial_name] = start_size + name_length - len(partial_name.encode("utf-8"))
             self._write_utf(labels[count])
 
         # this is the end of a name
         self._write_byte(0)
 
-    def _write_question(self, question: DNSQuestion) -> bool:
+    def _write_link_to_name(self, index: int_) -> None:
+        # If part of the name already exists in the packet,
+        # create a pointer to it
+        self._write_byte((index >> 8) | 0xC0)
+        self._write_byte(index & 0xFF)
+
+    def _write_question(self, question: DNSQuestion_) -> bool:
         """Writes a question to the packet"""
-        start_data_length, start_size = len(self.data), self.size
+        start_data_length = len(self.data)
+        start_size = self.size
         self.write_name(question.name)
         self.write_short(question.type)
         self._write_record_class(question)
         return self._check_data_limit_or_rollback(start_data_length, start_size)
 
-    def _write_record_class(self, record: Union[DNSQuestion, DNSRecord]) -> None:
+    def _write_record_class(self, record: Union[DNSQuestion_, DNSRecord_]) -> None:
         """Write out the record class including the unique/unicast (QU) bit."""
-        if record.unique and self.multicast:
-            self.write_short(record.class_ | _CLASS_UNIQUE)
+        class_ = record.class_
+        if record.unique is True and self.multicast:
+            self.write_short(class_ | _CLASS_UNIQUE)
         else:
-            self.write_short(record.class_)
+            self.write_short(class_)
 
-    def _write_ttl(self, record: DNSRecord, now: float) -> None:
+    def _write_ttl(self, record: DNSRecord_, now: float_) -> None:
         """Write out the record ttl."""
         self._write_int(record.ttl if now == 0 else record.get_remaining_ttl(now))
 
-    def _write_record(self, record: DNSRecord, now: float) -> bool:
+    def _write_record(self, record: DNSRecord_, now: float_) -> bool:
         """Writes a record (answer, authoritative answer, additional) to
         the packet.  Returns True on success, or False if we did not
         because the packet because the record does not fit."""
-        start_data_length, start_size = len(self.data), self.size
+        start_data_length = len(self.data)
+        start_size = self.size
         self.write_name(record.name)
         self.write_short(record.type)
         self._write_record_class(record)
@@ -308,13 +339,15 @@ class DNSOutgoing:
         self.write_short(0)  # Will get replaced with the actual size
         record.write(self)
         # Adjust size for the short we will write before this record
-        length = sum(len(d) for d in self.data[index + 1 :])
+        length = 0
+        for d in self.data[index + 1 :]:
+            length += len(d)
         # Here we replace the 0 length short we wrote
         # before with the actual length
         self._replace_short(index, length)
         return self._check_data_limit_or_rollback(start_data_length, start_size)
 
-    def _check_data_limit_or_rollback(self, start_data_length: int, start_size: int) -> bool:
+    def _check_data_limit_or_rollback(self, start_data_length: int_, start_size: int_) -> bool:
         """Check data limit, if we go over, then rollback and return False."""
         len_limit = _MAX_MSG_ABSOLUTE if self.allow_long else _MAX_MSG_TYPICAL
         self.allow_long = False
@@ -322,16 +355,22 @@ class DNSOutgoing:
         if self.size <= len_limit:
             return True
 
-        log.debug("Reached data limit (size=%d) > (limit=%d) - rolling back", self.size, len_limit)
+        if LOGGING_IS_ENABLED_FOR(LOGGING_DEBUG):  # pragma: no branch
+            log.debug(
+                "Reached data limit (size=%d) > (limit=%d) - rolling back",
+                self.size,
+                len_limit,
+            )
         del self.data[start_data_length:]
         self.size = start_size
 
-        rollback_names = [name for name, idx in self.names.items() if idx >= start_size]
+        start_size_int = start_size
+        rollback_names = [name for name, idx in self.names.items() if idx >= start_size_int]
         for name in rollback_names:
             del self.names[name]
         return False
 
-    def _write_questions_from_offset(self, questions_offset: int) -> int:
+    def _write_questions_from_offset(self, questions_offset: int_) -> int:
         questions_written = 0
         for question in self.questions[questions_offset:]:
             if not self._write_question(question):
@@ -339,7 +378,7 @@ class DNSOutgoing:
             questions_written += 1
         return questions_written
 
-    def _write_answers_from_offset(self, answer_offset: int) -> int:
+    def _write_answers_from_offset(self, answer_offset: int_) -> int:
         answers_written = 0
         for answer, time_ in self.answers[answer_offset:]:
             if not self._write_record(answer, time_):
@@ -347,7 +386,7 @@ class DNSOutgoing:
             answers_written += 1
         return answers_written
 
-    def _write_records_from_offset(self, records: Sequence[DNSRecord], offset: int) -> int:
+    def _write_records_from_offset(self, records: Sequence[DNSRecord], offset: int_) -> int:
         records_written = 0
         for record in records[offset:]:
             if not self._write_record(record, 0):
@@ -356,7 +395,11 @@ class DNSOutgoing:
         return records_written
 
     def _has_more_to_add(
-        self, questions_offset: int, answer_offset: int, authority_offset: int, additional_offset: int
+        self,
+        questions_offset: int_,
+        answer_offset: int_,
+        authority_offset: int_,
+        additional_offset: int_,
     ) -> bool:
         """Check if all questions, answers, authority, and additionals have been written to the packet."""
         return (
@@ -375,24 +418,20 @@ class DNSOutgoing:
         will be written out to a single oversized packet no more than
         _MAX_MSG_ABSOLUTE in length (and hence will be subject to IP
         fragmentation potentially)."""
-        return self._packets()
+        packets_data = self.packets_data
 
-    def _packets(self) -> List[bytes]:
-        if self.state == State.finished:
-            return self.packets_data
+        if self.state == STATE_FINISHED:
+            return packets_data
 
         questions_offset = 0
         answer_offset = 0
         authority_offset = 0
         additional_offset = 0
         # we have to at least write out the question
-        first_time = True
-        debug_enable = log.isEnabledFor(logging.DEBUG)
+        debug_enable = LOGGING_IS_ENABLED_FOR(LOGGING_DEBUG) is True
+        has_more_to_add = True
 
-        while first_time or self._has_more_to_add(
-            questions_offset, answer_offset, authority_offset, additional_offset
-        ):
-            first_time = False
+        while has_more_to_add:
             if debug_enable:
                 log.debug(
                     "offsets = questions=%d, answers=%d, authorities=%d, additionals=%d",
@@ -414,6 +453,8 @@ class DNSOutgoing:
             authorities_written = self._write_records_from_offset(self.authorities, authority_offset)
             additionals_written = self._write_records_from_offset(self.additionals, additional_offset)
 
+            made_progress = bool(self.data)
+
             self._insert_short_at_start(additionals_written)
             self._insert_short_at_start(authorities_written)
             self._insert_short_at_start(answers_written)
@@ -432,11 +473,14 @@ class DNSOutgoing:
                     additional_offset,
                 )
 
-            if self.is_query() and self._has_more_to_add(
+            has_more_to_add = self._has_more_to_add(
                 questions_offset, answer_offset, authority_offset, additional_offset
-            ):
+            )
+
+            if has_more_to_add and self.is_query():
                 # https://datatracker.ietf.org/doc/html/rfc6762#section-7.2
-                log.debug("Setting TC flag")
+                if debug_enable:  # pragma: no branch
+                    log.debug("Setting TC flag")
                 self._insert_short_at_start(self.flags | _FLAGS_TC)
             else:
                 self._insert_short_at_start(self.flags)
@@ -446,13 +490,18 @@ class DNSOutgoing:
             else:
                 self._insert_short_at_start(self.id)
 
-            self.packets_data.append(b''.join(self.data))
-            self._reset_for_next_packet()
+            packets_data.append(b"".join(self.data))
 
-            if (questions_written + answers_written + authorities_written + additionals_written) == 0 and (
-                len(self.questions) + len(self.answers) + len(self.authorities) + len(self.additionals)
-            ) > 0:
+            if not made_progress:
+                # Generating an empty packet is not a desirable outcome, but currently
+                # too many internals rely on this behavior.  So, we'll just return an
+                # empty packet and log a warning until this can be refactored at a later
+                # date.
                 log.warning("packets() made no progress adding records; returning")
                 break
-        self.state = State.finished
-        return self.packets_data
+
+            if has_more_to_add:
+                self._reset_for_next_packet()
+
+        self.state = STATE_FINISHED
+        return packets_data
